@@ -31,10 +31,10 @@ namespace GLib {
 		private IntPtr list_ptr = IntPtr.Zero;
 		private int length = -1;
 		private bool managed = false;
-		private bool elements_owned = false;
+		internal bool elements_owned = false;
 		protected System.Type element_type = null;
 
-                abstract internal IntPtr NthData (uint index);
+		abstract internal IntPtr NthData (uint index);
 		abstract internal IntPtr GetData (IntPtr current);
 		abstract internal IntPtr Next (IntPtr current);
 		abstract internal int Length (IntPtr list);
@@ -76,6 +76,11 @@ namespace GLib {
 			this.Append (Marshaller.StringToPtrGStrdup (item));
 		}
 
+		public void Append (object item)
+		{
+			this.Append (AllocNativeElement (item));
+		}
+
 		public void Prepend (IntPtr raw)
 		{
 			list_ptr = Prepend (list_ptr, raw);
@@ -113,7 +118,7 @@ namespace GLib {
 			object[] orig = new object[Count];
 			int i = 0;
 			foreach (object o in this)
-				orig[i++] = o;
+				orig [i++] = o;
 			
 			orig.CopyTo (array, index); 
 		}
@@ -122,6 +127,30 @@ namespace GLib {
 			private FilenameString () {}
 		}
 
+		IntPtr AllocNativeElement (object element)
+		{
+			if (element_type == null) {
+				if (element is IWrapper)
+					return (element as IWrapper).Handle;
+				else
+					return (IntPtr) GCHandle.Alloc (element);
+			} else {
+				if (element_type == typeof (string))
+					return Marshaller.StringToPtrGStrdup (element as string);
+				else if (element_type == typeof (FilenameString))
+					return Marshaller.StringToFilenamePtr (element as string);
+				else if (element_type == typeof (IntPtr))
+					return (IntPtr) GCHandle.Alloc (element);
+				else if (typeof (IWrapper).IsAssignableFrom (element_type))
+					return (element as IWrapper).Handle;
+				else if (element_type == typeof (int))
+					return new IntPtr ((int) element);
+				else if (element_type.IsValueType)
+					return Marshaller.StructureToPtrAlloc (element);
+			}
+			return IntPtr.Zero;
+		}
+
 		internal object DataMarshal (IntPtr data) 
 		{
 			object ret = null;
@@ -134,11 +163,17 @@ namespace GLib {
 					ret = data;
 				else if (element_type.IsSubclassOf (typeof (GLib.Object)))
 					ret = GLib.Object.GetObject (data, false);
+				else if (element_type.IsSubclassOf (typeof (GLib.Opaque)))
+					ret = GLib.Opaque.GetOpaque (data, element_type, elements_owned);
 				else if (element_type == typeof (int))
 					ret = (int) data;
 				else if (element_type.IsValueType)
 					ret = Marshal.PtrToStructure (data, element_type);
-				else
+				else if (element_type.IsInterface) {
+					Type adapter_type = element_type.Assembly.GetType (element_type.FullName + "Adapter");
+					System.Reflection.MethodInfo method = adapter_type.GetMethod ("GetObject", new Type[] {typeof(IntPtr), typeof(bool)});
+					ret = method.Invoke (null, new object[] {data, false});
+				} else
 					ret = Activator.CreateInstance (element_type, new object[] {data});
 
 			} else if (Object.IsObject (data))
@@ -150,7 +185,7 @@ namespace GLib {
 		[DllImport ("libglib-2.0-0.dll")]
 		static extern void g_free (IntPtr item);
 
-		[DllImport ("libglib-2.0-0.dll")]
+		[DllImport ("libgobject-2.0-0.dll")]
 		static extern void g_object_unref (IntPtr item);
 
 		public void Empty ()
@@ -159,7 +194,9 @@ namespace GLib {
 				for (uint i = 0; i < Count; i++)
 					if (typeof (GLib.Object).IsAssignableFrom (element_type))
 						g_object_unref (NthData (i));
-					else
+					else if (typeof (GLib.Opaque).IsAssignableFrom (element_type))
+						GLib.Opaque.GetOpaque (NthData (i), element_type, true).Dispose ();
+					else 
 						g_free (NthData (i));
 
 			if (managed)
